/*
 * libbb/smack_common.c
 *   -- common Smack utility functions
 *
 * Copyright 2007 Casey Schaufler <casey@schaufler-ca.com>
 */
#include "libbb.h"
#include <attr/xattr.h>

#define	SMACKPROCPATH	"/proc/%d/attr/current"
#define SMACKUSERFILE	"/etc/smack/user"
#define SMACKFILELINE	256

int smack_from_file(const char *path, char *result, int len, int follow)
{
	char buffer[SMACKBUFFSIZE];
	int rc;

	if (follow)
		rc = getxattr(path, SMACKATTR, buffer, SMACKBUFFSIZE - 1);
	else
		rc = lgetxattr(path, SMACKATTR, buffer, SMACKBUFFSIZE - 1);

	if (rc < 0)
		return rc;

	buffer[rc] = '\0';
	if (strlen(buffer) > len)
		return -EINVAL;

	strcpy(result, buffer);
	return rc;
}

int smack_from_proc(const int pid, char *result, int len)
{
	char path[SMACKFILELINE];
	char buffer[SMACKFILELINE];
	int fd;
	int i;

	sprintf(path, SMACKPROCPATH, (pid < 0) ? getpid() : pid);

	fd = open(path, O_RDONLY);
	if (fd < 0)
		return -EACCES;

	i = read(fd, buffer, SMACKFILELINE);
	close(fd);

	if (i <= 0 || i >= len)
		return -ERANGE;
	buffer[i] = '\0';
	strcpy(result, buffer);
	return 0;
}

int smack_to_file(const char *path, const char *smack, int follow)
{
	if (follow)
		return setxattr(path, SMACKATTR, smack, strlen(smack)+1, 0);
	return lsetxattr(path, SMACKATTR, smack, strlen(smack)+1, 0);
}

int smack_to_proc(const int pid, const char *smack)
{
	char path[SMACKFILELINE];
	int fd;
	int i;
	int slen;

	slen = strlen(smack) + 1;
	if (slen >= SMACKBUFFSIZE)
		return -ERANGE;

	sprintf(path, SMACKPROCPATH, (pid < 0) ? getpid() : pid);

	fd = open(path, O_RDWR);
	if (fd < 0)
		return -EACCES;

	i = write(fd, smack, slen);
	close(fd);

	if (i != slen)
		return -EINVAL;
	return 0;
}

int smack_user_default(const char *user, char *result, int len)
{
	char line[SMACKFILELINE];
	char ruser[SMACKFILELINE];
	char rsmack[SMACKFILELINE];
	FILE *fp;

	fp = fopen(bb_path_smack_user, "r");
	if (fp == NULL)
		return -ENOENT;

	while (fgets(line, SMACKFILELINE, fp) != NULL) {
		if (line[0] == '#' || line[0] == '\n')
			continue;
		if (sscanf(line, "%s %s", ruser, rsmack) != 2)
			continue;
		if (strcmp(ruser, user) != 0)
			continue;
		if (strlen(rsmack) >= len)
			return -ERANGE;
		strcpy(result, rsmack);
		return 0;
	}
	return -ENOENT;
}

int smack_user_allowed(const char *user, const char *smack)
{
	char line[SMACKFILELINE];
	char *ruser;
	char *rsmack;
	FILE *fp;
	int rc = 1;

	fp = fopen(SMACKUSERFILE, "r");
	if (fp == NULL)
		return -ENOENT;

	while (rc == 1 && fgets(line, SMACKFILELINE, fp) != NULL) {
		if (line[0] == '#' || line[0] == '\n')
			continue;

		ruser = strtok(line, " \t\n");
		if (ruser == NULL)
			continue;

		if (strcmp(ruser, user) != 0)
			continue;

		while (rc == 1 && (rsmack = strtok(NULL, " \t\n")) != NULL) {
			if (strcmp(rsmack, smack) == 0 || strcmp(rsmack, "+") == 0)
				rc = 0;
		}
	}
	fclose(fp);
	return (rc == 1) ? -ENOENT : rc;
}

int smack_user_add(const char *user, const char *smack)
{
	char line[SMACKFILELINE];
	char *ruser;
	char *rsmack;
	FILE *fp;
	int rc = 0;

	fp = fopen(SMACKUSERFILE, "a+");
	if (fp == NULL)
		return -ENOENT;

	while (fgets(line, SMACKFILELINE, fp) != NULL) {
		if (line[0] == '#' || line[0] == '\n')
			continue;

		ruser = strtok(line, " \t\n");
		if (ruser == NULL)
			continue;

		if (strcmp(ruser, user) == 0) {
			rc = -EEXIST;
			break;
		}
	}

	if (rc == 0)
		fprintf(fp, "%s %s\n", user, smack);

	fclose(fp);
	return rc;
}
